/*
 * c5g_hdmi local test code
 *
 */


#include "terasic_includes.h"
#include "i2c.h"
#include "system.h"

#include "encoder.h"
#include "adv7513.h"

#define ADV7341_SALVE_ADDR7		0x2a
#define ADV7341_SALVE_ADDR8		((alt_u8)(ADV7341_SALVE_ADDR7) << 1)

#define ADV7511_SALVE_ADDR7		0x39										// 0x39)PD-low 0x3D)PD-high
#define ADV7511_SALVE_ADDR8		((alt_u8)(ADV7513_SALVE_ADDR7) << 1)		// 0x72)PD-low 0x7A)PD-high

#define ADV7513_SALVE_ADDR7		0x39										// 0x39)PD-low 0x3D)PD-high
#define ADV7513_SALVE_ADDR8		((alt_u8)(ADV7513_SALVE_ADDR7) << 1)		// 0x72)PD-low 0x7A)PD-high

#define ADV7513_EDID_ADDR8		0x7e

static int bin_dump(const alt_u8 *buf, int len)
{
	int r, c;
	const alt_u8 *p = buf;

	int rows = len / 16;
	int cols = len % 16;

	if (len <= 0)
		return 0;

	for (r = 0; r < rows; ++r, p += 16) {
		printf("%04x | %02x %02x %02x %02x %02x %02x %02x %02x - ",
				r * 16, p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7]);

		printf("%02x %02x %02x %02x %02x %02x %02x %02x\n",
				p[8], p[9], p[10], p[11], p[12], p[13], p[14], p[15]);
	}

	if (cols > 0) {
		printf("%04x | ", (rows - 1) * 16);
		for (c = 0; c < cols; ++c, ++p) {
			if (c == 8) {
				printf("- %02x ", *p);
			} else {
				printf("%02x ", *p);
			}
		}
		printf("\n");
	}

	return len;
}

#define HDMI_TX_INT_MASK			0x01

// warning: for simplicity, we handle-time consuming tasks this ISR ...
//  - use printf() (normally you should not)
//  - handle hot-plug events (usually it should be deferred to application level rather than in interrupt level)
static void adv7513_int_handler(void *context)
{
	struct hdmi_encoder *ec = (struct hdmi_encoder *)context;
	alt_u32 irq_status;
	alt_u8 regv;
	struct adv7513_int_mask int_mask;
	struct adv7513_int_status int_status;
	int r;

	// get system irq status
	irq_status = IORD_ALTERA_AVALON_PIO_EDGE_CAP(HDMI_TX_INT_N_BASE) & HDMI_TX_INT_MASK;
	//if (irq_status == 0)
	//	return;

	// clear system irq flag
	IOWR_ALTERA_AVALON_PIO_EDGE_CAP(HDMI_TX_INT_N_BASE, 0x00);

	// reload adv7513 int mask value
	adv7513_long_to_int_mask(&int_mask, ec->int_mask);

	// get adv7513 int status
	r = reg_read(ec, 0x93, &regv);
	if (r != 0)
		return;
	int_status.reg93 = regv;

	r = reg_read(ec, 0x96, &regv);
	if (r != 0)
		return;
	int_status.reg96 = regv;

	r = reg_read(ec, 0x97, &regv);
	if (r != 0)
		return;
	int_status.reg97 = regv;

	// disable adv7513 int
	reg_write(ec, 0x92, 0);
	reg_write(ec, 0x94, 0);
	reg_write(ec, 0x95, 0);

	// clear adv7513 int status
	if (int_status.reg93 != 0) {
		reg_write(ec, 0x93, 0xff);
	}

	if (int_status.reg96 != 0) {
		reg_write(ec, 0x96, 0xff);
	}

	if (int_status.reg97 != 0) {
		reg_write(ec, 0x97, 0xff);
	}

	// -------------------------------------------------------
	// handle adv7513 interrupt events
	// WARNING: this should be an application level task ...
	//
	if (int_status.reg96_bits.monitor_sense || int_status.reg96_bits.hpd) {

		alt_u8 edid[256];
		int dvi_mode = 0;
		int pwr_state = 0;
		int pwron_retry_count = 200;
		int m_sense_retry_count = 30;

		while (m_sense_retry_count-- > 0) {
			r = reg_read(ec, ADV7513_REG_STATUS, &regv);
			if (r != 0)
				goto __unplugged_iret;

			if ((regv & ((ADV7513_STATUS_HPD) | (ADV7513_STATUS_MONITOR_SENSE))) == ((ADV7513_STATUS_HPD) | (ADV7513_STATUS_MONITOR_SENSE)))
				break;
		}

		if ((regv & ((ADV7513_STATUS_HPD) | (ADV7513_STATUS_MONITOR_SENSE))) != ((ADV7513_STATUS_HPD) | (ADV7513_STATUS_MONITOR_SENSE)))
			goto __unplugged_iret;

		// now we have both HPD and monitor-sense signal asserted

		// power up the hdmi-encoder ...
		r = adv7513_do_hpd_powerup(ec);
		if (r != 0)
			goto __unplugged_iret;

		// readback edid data
		r = adv7513_edid_read(ec, edid, 0);
		if (r != 0)
			dvi_mode = 1;

		if (edid[0x7e] == 0)
			dvi_mode = 1;

		// configure the hdmi-encoder ...
		while (pwron_retry_count > 0) {
			r = adv7513_chip_init(ec, dvi_mode);
			if (r != 0)
				goto __unplugged_iret;

			r = adv7513_power_state(ec, &pwr_state);
			if (r != 0)
				goto __unplugged_iret;

			if (pwr_state != 0)
				break;
			--pwron_retry_count;
		}

		r = adv7513_power_state(ec, &pwr_state);
		if (r != 0)
			goto __unplugged_iret;

		if (pwr_state == 0)
			goto __unplugged_iret;

		goto __plugged_iret;
	}

__unplugged_iret:

	// enable interrupts
	reg_write(ec, 0x92, int_mask.reg92);
	reg_write(ec, 0x94, int_mask.reg94);
	reg_write(ec, 0x95, int_mask.reg95);

	return;

__plugged_iret:

	// enable interrupts
	reg_write(ec, 0x92, int_mask.reg92);
	reg_write(ec, 0x94, int_mask.reg94 & 0xbf);
	reg_write(ec, 0x95, int_mask.reg95);

	return;
}

int adv7513_int_setup(struct hdmi_encoder *ec, const struct adv7513_int_mask *mask, alt_isr_func isr)
{
	int r = 0;

	printf("adv7513_int_setup()\n");

	ec->int_mask = adv7513_int_mask_to_long(mask);

	// disable HDMI_TX_INT
	alt_ic_irq_disable(HDMI_TX_INT_N_IRQ_INTERRUPT_CONTROLLER_ID, HDMI_TX_INT_N_IRQ);

	r = reg_write(ec, 0x92, 0);
	if (r != 0)
		return -1;

	r = reg_write(ec, 0x94, 0);
	if (r != 0)
		return -1;

	r = reg_write(ec, 0x95, 0);
	if (r != 0)
		return -1;

	// clear int flags
	IOWR_ALTERA_AVALON_PIO_EDGE_CAP(HDMI_TX_INT_N_BASE, 0x00);

	r = reg_write(ec, 0x93, 0xff);
	if (r != 0)
		return -1;

	r = reg_write(ec, 0x96, 0xff);
	if (r != 0)
		return -1;

	r = reg_write(ec, 0x97, 0xff);
	if (r != 0)
		return -1;

	// enable interrupt
	if (ec->int_mask > 0) {
		// register isr
		if (isr) {
			// register interrupt handler
			r = alt_ic_isr_register(HDMI_TX_INT_N_IRQ_INTERRUPT_CONTROLLER_ID, HDMI_TX_INT_N_IRQ, isr, ec, NULL);
			if (r != 0)
				return -1;
		}

		// enable interrupts
		r = reg_write(ec, 0x92, mask->reg92);
		if (r != 0)
			return -1;

		r = reg_write(ec, 0x94, mask->reg94);
		if (r != 0)
			return -1;

		r = reg_write(ec, 0x95, mask->reg95);
		if (r != 0)
			return -1;

		IOWR_ALTERA_AVALON_PIO_IRQ_MASK(HDMI_TX_INT_N_BASE, HDMI_TX_INT_MASK); // 0x01: enable bit0

		alt_ic_irq_enable(HDMI_TX_INT_N_IRQ_INTERRUPT_CONTROLLER_ID, HDMI_TX_INT_N_IRQ);
	}

	return 0;
}


bool hdmi_tx_start(void)
{
	int r = 0, rr = 0;

	static struct hdmi_encoder encoder = {
		.slave_addr = ADV7513_SALVE_ADDR8,
		.edid_addr = ADV7513_EDID_ADDR8,
		.hpd = 0,
		.chip_id = {0},
		.chip_rev = {0},
	};


	struct adv7513_int_mask ie_mask;

	memset(&ie_mask, 0, sizeof(ie_mask));

	printf("\n===== c5g/de1-77 adv7513 hdmi encoder demo =====\n");

	r = adv7513_chip_identify(&encoder);
	if (r != 0) {
		printf("[err] failed to do chip-identification! (%d)\n", r);
		rr = -1;
		goto _safe_exit;
	}

	if ((encoder.chip_id[0] == 0x11) && (encoder.chip_id[1] == 0x75)) {
		printf("encoder chip : ADV%02x%02x rev: 0x%02x\n", encoder.chip_id[1], encoder.chip_id[0], encoder.chip_rev[0]);
	} else {
		printf("encoder chip : unknown (id: 0x%02x 0x%02x rev: 0x%02x)\n", encoder.chip_id[0], encoder.chip_id[1], encoder.chip_rev[0]);
	}

	ie_mask.reg94_bits.hpd = 1;
	ie_mask.reg94_bits.monitor_sense = 1;

	r = adv7513_int_setup(&encoder, &ie_mask, adv7513_int_handler);
	if (r != 0) {
		printf("[err] failed to do HDMI_TX_INT interrupt setup! (%d)\n", r);
		rr = -1;
		goto _safe_exit;
	} else {
		printf("success to setup HDMI_TX_INT interrupt handler.\n");
	}

	adv7513_kick_up(&encoder);


_safe_exit:

	printf("\nprogram exit!\n");

	return rr;
}

