/*
 * main.c
 *
 *  Created on: 2013/7/17
 *      Author: Richard
 */
#include "terasic_includes.h"
#include "vip_fr.h"
#include "simple_graphics.h"
#include "simple_text.h"

#define USE_SMALL_FONT  // enable for higher frame rate


bool hdmi_tx_start(void);


#define FR_WIDTH			1024
#define FR_HEIGHT	 		768
#define FR_BYTES_PER_PIXEL	 4
#define FR_SIZE             (FR_WIDTH*FR_HEIGHT*FR_BYTES_PER_PIXEL)

#define FR_FRAME_0 	   (MEM_IF_LPDDR2_EMIF_BASE + 0x00000000)
#define FR_FRAME_1     (FR_FRAME_0 + FR_SIZE)

#define KEY_TRIGGER_FREE	0x01
#define KEY_RESET_MAX_MIN	0x02
#define SEARCH_WIDTH		800

//const float expected_adc_value[8] = {4.0, 1.65, 3.0, 2.5, 2.0, 1.5,  1.0, 0.5};
void ADC_Waveform(VIP_FRAME_READER *pFrameReader){
#ifdef USE_SMALL_FONT
	const int Y_GAP = 25;
#else
	const int Y_GAP = 40;
#endif

	int nNum, i, x, y;
	int nActiveChannel;
	alt_u16 szAdcValue[1024+SEARCH_WIDTH];
	alt_u32 AdcMax=0, AdcMin=5000, AdcAvg;
	int VolMax = 5000;
	char szText[256];
	const int X_GAP = 10;
	int nPreCh = -1;
	int radius = Y_GAP/3;
	bool radus_add = TRUE;
	alt_u32 time_start, time_elapsed;
	float fRate;
	void *pFrame;
	bool bFreeRun = TRUE;
	int KeyStatus, PreKeyStatus = 0x00;
	bool bTriggerValueReady = FALSE;
	int nTriggerValue = -1;
	int nStartX, CurMax, CurMini, nDataIndex;
	const int nTriggerMiniHeight = 300;




	nNum = sizeof(szAdcValue)/sizeof(szAdcValue[0]);
	printf("screen W=%d, H=%d\r\n", pFrameReader->width, pFrameReader->height);

	while(1){
		time_start = alt_nticks();


		// specify active channel
		nActiveChannel = IORD(SW_BASE,0) & 0x07;

		// reset max/min ?
		KeyStatus = ~IORD(KEY_BASE, 0x00) & 0x0F;
		if ((KeyStatus & KEY_RESET_MAX_MIN) || (nActiveChannel != nPreCh)){
			// reset Max & Min
			AdcMax=0;
			AdcMin=5000;
		}

		// toggle trigger/free-run
		if ((KeyStatus & KEY_TRIGGER_FREE) && !(PreKeyStatus & KEY_TRIGGER_FREE)){
			bFreeRun = bFreeRun?FALSE:TRUE;
			if (!bFreeRun)
				bTriggerValueReady = FALSE;
			printf("Mode=%s\r\n",bFreeRun?"Free Run": "Auto Trigger");
		}

		if (!bFreeRun && (nActiveChannel != nPreCh)){
			bTriggerValueReady = FALSE; // find trigger point again
		}

		nPreCh = nActiveChannel;
		PreKeyStatus = KeyStatus;

		IOWR(LED_BASE, 0x00, bFreeRun?0x00:0x01); // LED0 light: trigger mode
		/////////////////////////////
		// read adc
		if (!ADC_LTC2308_Read(nActiveChannel, nNum, szAdcValue)){
			printf("failed to read adc\r\n");
			continue;
		}

		// finding trigger value if in trigger mode
		if (!bFreeRun && !bTriggerValueReady){
			CurMax = 0;
			CurMini = 5000;
			for(i=X_GAP;i<nNum;i++){
				if (szAdcValue[i] > CurMax)
					CurMax = szAdcValue[i];
				if (szAdcValue[i] < CurMini)
					CurMini = szAdcValue[i];
			}
			if ((CurMax - CurMini) > nTriggerMiniHeight){
				// find trigger point
				bTriggerValueReady = TRUE;
				nTriggerValue = (CurMax-CurMini)/2;
				printf("Max=%d, mini=%d, TriggerValue=%d\r]\n", CurMax, CurMini, nTriggerValue);
			}
		}

		/////////////////////////////
		// clear canvas
		vid_screen_clear(pFrameReader);




		/////////////////////////////
		// Title with green background
		vid_draw_box (0, 0, pFrameReader->width-1, Y_GAP, BLUE_24, DO_FILL, pFrameReader);
		sprintf(szText, "Terasic ADC Demo");
#ifdef USE_SMALL_FONT
		vid_print_string(pFrameReader->width/2  + 0 - 120,  6, WHITE_24,   cour10_font, pFrameReader, szText);
#else
		//vid_print_string_alpha(pFrameReader->width/2  + 2 - 170,  2+2  , GRAY_24,  BLACK_24, tahomabold_32, pFrameReader, szText);
		//vid_print_string_alpha(pFrameReader->width/2  + 0 - 170,  2+0, WHITE_24,   BLACK_24, tahomabold_32, pFrameReader, szText);
		vid_print_string_no_alpha(pFrameReader->width/2  + 0 - 170,  2+0, WHITE_24,   tahomabold_32, pFrameReader, szText);
#endif
		/////////////////////////////
		// draw 0V ref
		vid_draw_line(0, pFrameReader->height - Y_GAP, pFrameReader->width-1, pFrameReader->height - Y_GAP, 1, GRAY_24, pFrameReader);
		sprintf(szText, "0.0V");

#ifdef USE_SMALL_FONT
		vid_print_string(X_GAP, pFrameReader->height - Y_GAP - 12,  GRAY_24,   cour10_font, pFrameReader, szText);
#else
		vid_print_string_alpha(X_GAP, pFrameReader->height - Y_GAP - 26  , GRAY_24,  BLACK_24, tahomabold_20, pFrameReader, szText);
#endif

		/////////////////////////////
		// show animation mark

		vid_draw_circle(Y_GAP/2, Y_GAP/2, radius, WHITE_24, DO_NOT_FILL, pFrameReader);
		vid_draw_circle(pFrameReader->width - Y_GAP/2, Y_GAP/2, radius, WHITE_24, DO_NOT_FILL, pFrameReader);
		if (radus_add){
			radius ++;
			if (radius >= Y_GAP/3){
				radius = Y_GAP/3;
				radus_add = FALSE;
			}

		}else{
			radius--;
			if (radius == 0){
				radius = 1;
				radus_add = TRUE;
			}
		}

		/////////////////////////////
		// draw wave
		nStartX = X_GAP;
		if (!bFreeRun && bTriggerValueReady){
			// find start X;
			for(i=X_GAP;i<SEARCH_WIDTH;i++){
				if (szAdcValue[i-3] < nTriggerValue && szAdcValue[i+3] >  nTriggerValue){
					nStartX = i;
					break;
				}
			}

		}

//		for(i=nStartX;i<pFrameReader->width-nStartX;i++){
		for(i=X_GAP;i<pFrameReader->width-X_GAP;i++){
			nDataIndex = nStartX+i-X_GAP;
			x = i;
			y = pFrameReader->height - Y_GAP;
			y -= szAdcValue[nDataIndex]*(pFrameReader->height-2*Y_GAP)/VolMax;
			vid_set_pixel(x, y, YELLOW_24, pFrameReader);
		}

		vid_draw_box (0, pFrameReader->height-Y_GAP+2, pFrameReader->width-1, pFrameReader->height-1, GRAY_24, DO_FILL, pFrameReader);
		/////////////////////////////
		// show voltage info. min/max/avg
		AdcAvg = 0;
		for(i=X_GAP;i<pFrameReader->width-X_GAP;i++){
			nDataIndex = nStartX+i-X_GAP;
			if (szAdcValue[nDataIndex] > AdcMax)
				AdcMax = szAdcValue[nDataIndex];
			if (szAdcValue[nDataIndex] < AdcMin)
				AdcMin = szAdcValue[nDataIndex];
			AdcAvg += szAdcValue[nDataIndex];
		}
		AdcAvg /= nNum;
		sprintf(szText, "CH=%d, Max=%.3fV, Min=%.3fV, Avg=%.3fV", nActiveChannel, (float)AdcMax/1000.0, (float)AdcMin/1000.0, (float)AdcAvg/1000.0);
#ifdef USE_SMALL_FONT
		vid_print_string(X_GAP, pFrameReader->height-Y_GAP+8,  YELLOW_24, cour10_font, pFrameReader, szText);
#else
		vid_print_string_alpha(X_GAP, pFrameReader->height-Y_GAP+1,  YELLOW_24, BLACK_24, tahomabold_20, pFrameReader, szText);
#endif
		//vid_print_string_no_alpha(0, pFrameReader->height-Y_GAP+1,  YELLOW_24, tahomabold_20, pFrameReader, szText);

		// menu

		if (bFreeRun){
		  sprintf(szText, "SW[2:0]:Channel, KEY1: Reset Max/Min, KEY0:Auto-Trigger/*Free-Run");
		}else{
			if (!bTriggerValueReady)
				sprintf(szText, "SW[2:0]:Channel, KEY1: Reset Max/Min, KEY0:*Auto-Trigger/Free-Run");
			else
				  sprintf(szText, "SW[2:0]:Channel, KEY1: Reset Max/Min, KEY0:*Auto-Trigger(%.3fV)/Free-Run", (float)nTriggerValue/1000.0);
		}
#ifdef USE_SMALL_FONT
		vid_print_string(pFrameReader->width-600, pFrameReader->height-Y_GAP+8,  WHITE_24, cour10_font, pFrameReader, szText);
#else
		vid_print_string_alpha(pFrameReader->width-520, pFrameReader->height-Y_GAP+1,  GRAY_24, BLACK_24, tahomabold_20, pFrameReader, szText);
		//vid_print_string_no_alpha(pFrameReader->width-460, pFrameReader->height-Y_GAP+1,  GRAY_24, tahomabold_20, pFrameReader, szText);
#endif

		/////////////////////////////
		// update canvas
		VIPFR_ActiveDrawFrame(pFrameReader);


		// calculate frame rate
		time_elapsed = alt_nticks() - time_start;
		fRate = (float)alt_ticks_per_second() / (float)time_elapsed;
		printf("FPS=%.3f\r\n", fRate);



	}

}


int main(int argc, const char *argv[]){
	VIP_FRAME_READER *pFrameReader;


	IOWR(LED_BASE, 0x00, 0xFF); // turn on green led

	hdmi_tx_start();


	pFrameReader = VIPFR_Init(ALT_VIP_VFR_0_BASE, (void *)FR_FRAME_0, (void *)FR_FRAME_1, FR_WIDTH, FR_HEIGHT);
	if (pFrameReader){
		VIPFR_Go(pFrameReader, TRUE);
		ADC_Waveform(pFrameReader);
		//Moving_Box(pFrameReader);


	}else{
		printf("failed to init frame reader");
	}


   while (1);
	return 0;

}

